import cuqi
import numpy as np
import scipy
import pytest
from packaging import version

def test_RegularizedGaussian_default_init():
    """ Test that the implicit regularized Gaussian requires at least 1 regularization argument """

    with pytest.raises(ValueError, match="At least some "):
        x = cuqi.implicitprior.RegularizedGaussian(np.zeros(5), 1)

def test_RegularizedGaussian_guarding_statements():
    """ Test that we catch incorrect initialization of RegularizedGaussian """

    # More than 1 argument
    with pytest.raises(ValueError, match="User-defined proximals and "):
        cuqi.implicitprior.RegularizedGaussian(np.zeros(5), 1, proximal=lambda s,z: s, constraint="nonnegativity")

    # Proximal
    with pytest.raises(ValueError, match="Proximal needs to be callable or a list. See documentation."):
        cuqi.implicitprior.RegularizedGaussian(np.zeros(5), 1, proximal=1)

    with pytest.raises(ValueError, match="Proximal should take 2 arguments"):
        cuqi.implicitprior.RegularizedGaussian(np.zeros(5), 1, proximal=lambda s: s)

    # Projector
    with pytest.raises(ValueError, match="Projector needs to be callable"):
        cuqi.implicitprior.RegularizedGaussian(np.zeros(5), 1, projector=1)

    with pytest.raises(ValueError, match="Projector should take 1 argument"):
        cuqi.implicitprior.RegularizedGaussian(np.zeros(5), 1, projector=lambda s,z: s)

def test_creating_restorator():
    """ Test creating the object from restorator class."""

    def func(x, restoration_strength=0.1):
        return x, True
    restorator = cuqi.implicitprior.RestorationPrior(func)
    assert np.allclose(restorator.restore(np.ones(4), 0.1), np.ones(4))
    assert restorator.info == True

def test_handling_invalid_restorator():
    """ Test handling invalid restorator."""
    # Invalid return type 1: None
    def func_1(x, restoration_strength=0.1):
        return
    restore_prior_1 = cuqi.implicitprior.RestorationPrior(func_1)
    with pytest.raises(ValueError, match=r"Unsupported return type .*"):
        restore_prior_1.restore(np.ones(4), 0.1)
    # Invalid return type 2: one parameter
    def func_2(x, restoration_strength=0.1):
        return x
    restore_prior_2 = cuqi.implicitprior.RestorationPrior(func_2)
    with pytest.raises(ValueError, match=r"Unsupported return type .*"):
        restore_prior_2.restore(np.ones(4), 0.1)
    # Invalid return type 3: tuple with 3 elements
    def func_3(x, restoration_strength=0.1):
        return x, None, False
    restore_prior_3 = cuqi.implicitprior.RestorationPrior(func_3)
    with pytest.raises(ValueError, match=r"Unsupported return type .*"):
        restore_prior_3.restore(np.ones(4), 0.1)
    # Invalid return type 4: list with 2 elements
    def func_4(x, restoration_strength=0.1):
        return [x, None]
    restore_prior_4 = cuqi.implicitprior.RestorationPrior(func_4)
    with pytest.raises(ValueError, match=r"Unsupported return type .*"):
        restore_prior_4.restore(np.ones(4), 0.1)

def test_creating_restorator_with_potential():
    """ Test creating the object from restorator class with a potential."""

    def func(x, restoration_strength=1):
        return x/(1+restoration_strength), True
    def potential(x):
        return (x**2).sum()/2
    restorator = cuqi.implicitprior.RestorationPrior(restorator=func, potential=potential)
    assert np.allclose(restorator.restore(np.ones(1), restoration_strength=1), np.ones(1)/(1+1))
    assert restorator.info == True
    assert restorator.logpdf(np.ones(4)) == -2
    

def test_creating_moreau_yoshida_prior_gradient():
    """ Test creating MoreauYoshidaPrior."""

    def func(x, restoration_strength=1):
        return x/(1+restoration_strength), True
    def potential(x):
        return (x**2).sum()/2
    restorator = cuqi.implicitprior.RestorationPrior(func,
                                                    potential=potential)
    myprior = cuqi.implicitprior.MoreauYoshidaPrior(restorator, smoothing_strength=0.1)
    assert np.allclose(myprior.gradient(np.ones(1)), -np.ones(1)/(1+myprior.smoothing_strength))
    assert myprior.logpdf(np.ones(1)) == -0.5*myprior.smoothing_strength/(1+myprior.smoothing_strength)
            
def test_ConstrainedGaussian_alias():
    """ Test that the implicit constrained Gaussian is a correct allias for an implicit regularized Gaussian """

    x = cuqi.implicitprior.ConstrainedGaussian(np.zeros(5), 1, constraint="nonnegativity")

    assert isinstance(x, cuqi.implicitprior.RegularizedGaussian)
    assert x.preset["constraint"] == "nonnegativity"
    assert x.preset["regularization"] is None

def test_NonnegativeGaussian_alias():
    """ Test that the implicit nonnegative Gaussian is a correct allias for an implicit regularized Gaussian """

    x = cuqi.implicitprior.NonnegativeGaussian(np.zeros(5), 1)

    assert isinstance(x, cuqi.implicitprior.RegularizedGaussian)
    assert x.preset["constraint"] == "nonnegativity"
    assert x.preset["regularization"] is None

def test_ConstrainedGMRF_alias():
    """ Test that the implicit constrained GMRF is a correct allias for an implicit regularized GMRF """

    x = cuqi.implicitprior.ConstrainedGMRF(np.zeros(5), 1, constraint="nonnegativity")

    assert isinstance(x, cuqi.implicitprior.RegularizedGMRF)
    assert x.preset["constraint"] == "nonnegativity"
    assert x.preset["regularization"] is None

def test_NonnegativeGMRF_alias():
    """ Test that the implicit nonnegative GMRF is a correct allias for an implicit regularized GMRF """

    x = cuqi.implicitprior.NonnegativeGMRF(np.zeros(5), 1)

    assert isinstance(x, cuqi.implicitprior.RegularizedGMRF)
    assert x.preset["constraint"] == "nonnegativity"
    assert x.preset["regularization"] is None

def test_RegularizedUnboundedUniform_is_RegularizedGaussian():
    """ Test that the implicit regularized unbounded uniform create a Regularized Gaussian with zero sqrtprec """
    # NOTE: Test is based on the current assumption that the regularized uniform is modeled as a Gaussian with zero precision. This might change in the future.

    x = cuqi.implicitprior.RegularizedUnboundedUniform(cuqi.geometry.Continuous1D(5), regularization="l1", strength = 5.0)
    
    assert np.allclose(x.gaussian.sqrtprec, 0.0)

def test_RegularizedGaussian_conditioning_constrained():
    """ Test that conditioning the implicit regularized Gaussian works as expected """
    
    x  = cuqi.implicitprior.RegularizedGMRF(lambda a:a*np.ones(2**2),
                     prec = lambda b:5*b,
                     constraint = "nonnegativity",
                     geometry = cuqi.geometry.Image2D((2,2)))
    
    assert x.get_mutable_variables() == ['mean', 'prec']
    assert x.get_conditioning_variables() == ['a', 'b']

    x = x(a=1, b=2)

    assert np.allclose(x.mean, [1, 1, 1, 1])
    assert np.allclose(x.prec, 10)

def test_RegularizedGaussian_conditioning_strength():
    """ Test that conditioning the implicit regularized Gaussian works as expected """
    
    x  = cuqi.implicitprior.RegularizedGMRF(lambda a:a*np.ones(2**2),
                     prec = lambda b:5*b,
                     regularization = "tv",
                     strength = lambda c:c*2,
                     geometry = cuqi.geometry.Image2D((2,2)))
    
    assert x.get_mutable_variables() == ['mean', 'prec', 'strength']
    assert x.get_conditioning_variables() == ['a', 'b', 'c']

    x = x(a=1, b=2, c=3)

    assert np.allclose(x.mean, [1, 1, 1, 1])
    assert np.allclose(x.prec, 10)
    assert np.allclose(x.strength, 6)

def test_RegularizedGaussian_double_preset():
    """ Test that the implicit RegularizedGaussian can handle combined regularization and constraint presets """

    constraint = "nonnegativity"
    regularization = "tv"
    x = cuqi.implicitprior.RegularizedGaussian(np.zeros(5), 1,
                                                regularization = regularization, strength = 5,
                                                constraint = constraint)

    # Check that the correct presets are set
    assert x.preset["constraint"] == constraint
    assert x.preset["regularization"] == regularization
    # Check whether the constructed proximal list is of correct size
    assert len(x.proximal) == 2
    assert len(x.proximal[0]) == 2
    assert len(x.proximal[1]) == 2

@pytest.mark.skipif(version.parse(scipy.__version__) < version.parse('1.12.0'), reason="isotonic_regression in not available in scipy < 1.12.0")
def test_regression_increasing():
    """ Regression test for the increasing constraints in RegularizedGaussian"""

    np.random.seed(1337)

    n = 5
    A, _, _ = cuqi.testproblem.Deconvolution1D(dim=n, phantom='square').get_components()

    y_obs = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

    
    d = cuqi.distribution.Gamma(1, 1e-4)
    x = cuqi.implicitprior.RegularizedGaussian(mean = np.zeros(n), prec = lambda d : d,
                            constraint = "increasing",
                            geometry = A.domain_geometry,)
    y = cuqi.distribution.Gaussian(A@x, prec = 1000)

    joint = cuqi.distribution.JointDistribution(d, y, x)
    posterior = joint(y=y_obs)
   
    sampling_strategy = {
        'd': cuqi.experimental.mcmc.Conjugate(),
        'x': cuqi.experimental.mcmc.RegularizedLinearRTO(maxit=50, penalty_parameter=20, adaptive = False),
                        }
    sampler = cuqi.experimental.mcmc.HybridGibbs(posterior, sampling_strategy)

    sampler.sample(10)
    samples = sampler.get_samples()
    
    assert np.allclose(samples['x'].samples[:,-1], [0.07151248, 0.08253823, 0.3236245, 0.44846062, 0.61276932])
  
def test_regression_convex():
    """ Regression test for the convexity constraints in RegularizedGaussian"""

    np.random.seed(1337)

    n = 5
    A, _, _ = cuqi.testproblem.Deconvolution1D(dim=n, phantom='square').get_components()

    y_obs = np.array([1, 0, -1, 2, 6])

    
    d = cuqi.distribution.Gamma(1, 1e-4)
    x = cuqi.implicitprior.RegularizedGaussian(mean = np.zeros(n), prec = lambda d : d,
                            constraint = "convex",
                            geometry = A.domain_geometry,)
    y = cuqi.distribution.Gaussian(A@x, prec = 1000)

    joint = cuqi.distribution.JointDistribution(d, y, x)
    posterior = joint(y=y_obs)
   
    sampling_strategy = {
        'd': cuqi.experimental.mcmc.Conjugate(),
        'x': cuqi.experimental.mcmc.RegularizedLinearRTO(maxit=50, penalty_parameter=20, adaptive = False)
                        }
    sampler = cuqi.experimental.mcmc.HybridGibbs(posterior, sampling_strategy)

    sampler.sample(10)
    samples = sampler.get_samples()
    
    assert np.allclose(samples['x'].samples[:,-1], [-323.91409888, -194.50250442, -35.74426677, 167.01548949, 395.18988251])
  



# Test logic in posterior class. In particular when some distributions have default geometry

import pytest
import numpy as np
import cuqi
from .test_model import MultipleInputTestModel

def test_posterior_should_give_Model_priority_for_geometry():
    A = cuqi.model.Model(lambda x: x.ravel(), range_geometry=10**2, domain_geometry=cuqi.geometry.Image2D((10, 10)))
    x = cuqi.distribution.Gaussian(0, 1, geometry=10**2) # Default geometry
    y = cuqi.distribution.Gaussian(A(x), 1)
    posterior = cuqi.distribution.Posterior(y.to_likelihood(np.zeros(10**2)), x)
    assert type(posterior.geometry) == cuqi.geometry.Image2D
    assert posterior.geometry == A.domain_geometry

def test_posterior_should_give_prior_priority_for_geometry_if_default_model_geometry():
    A = cuqi.model.Model(lambda x: x.ravel(), range_geometry=10**2, domain_geometry=(10,10))
    x = cuqi.distribution.Gaussian(0, 1, geometry=cuqi.geometry.Image2D((10, 10)))
    y = cuqi.distribution.Gaussian(A(x), 1)
    posterior = cuqi.distribution.Posterior(y.to_likelihood(np.zeros(10**2)), x)
    assert type(posterior.geometry) == cuqi.geometry.Image2D
    assert posterior.geometry == x.geometry

def test_posterior_should_give_likelihood_priority_for_geometry_if_default_prior_geometry():
    x = cuqi.distribution.Gaussian(0, 1, geometry=(10,10))
    y = cuqi.likelihood.UserDefinedLikelihood(logpdf_func=lambda x: x, geometry=cuqi.geometry.Image2D((10, 10)))
    posterior = cuqi.distribution.Posterior(y, x)
    assert type(posterior.geometry) == cuqi.geometry.Image2D
    assert posterior.geometry == y.geometry

def test_posterior_should_not_allow_prior_wrong_shape_if_model_has_geometry():
    A = cuqi.model.Model(lambda x: x.ravel(), range_geometry=10**2, domain_geometry=cuqi.geometry.Image2D((10, 10)))
    x = cuqi.distribution.Gaussian(0, 1, geometry=cuqi.geometry.Continuous1D(10**2)) #1D shape
    y = cuqi.distribution.Gaussian(A(x), 1)

    with pytest.raises(ValueError):
        posterior = cuqi.distribution.Posterior(y.to_likelihood(np.zeros(10**2)), x)


def test_joint_distribution_with_multiple_inputs_model_reduce_to_posterior():
    """Test that a joint distribution based on a model with multiple inputs
    reduces to a posterior when data is observed and all inputs (except 1) are
    specified."""

    test_model = MultipleInputTestModel.helper_build_three_input_test_model()
    model = cuqi.model.Model(
        test_model.forward_map,
        gradient=test_model.gradient_form2,
        domain_geometry=test_model.domain_geometry,
        range_geometry=test_model.range_geometry,
    )

    # Create priors
    x_dist = cuqi.distribution.Gaussian(
        mean=np.zeros(3),
        cov=np.eye(3))
    y_dist = cuqi.distribution.Gaussian(
        mean=np.zeros(2),
        cov=np.eye(2))
    z_dist = cuqi.distribution.Gaussian(
        mean=np.zeros(3),
        cov=np.eye(3))

    # Create data distribution
    data_dist = cuqi.distribution.Gaussian(
        mean=model(x_dist, y_dist, z_dist), cov = 1.0)

    # Create likelihood
    likelihood = data_dist(data_dist = np.array([2,2,3]))

    # Create joint distribution
    posterior = cuqi.distribution.JointDistribution(
        likelihood,
        x_dist,
        y_dist,
        z_dist
    )

    assert isinstance(posterior, cuqi.distribution.JointDistribution)

    # This should reduce to a posterior object
    post_x_y = posterior(x_dist=np.array([1, 1, 1]), y_dist=np.array([0, 1]))

    assert isinstance(post_x_y, cuqi.distribution.Posterior)


from __future__ import annotations
from typing import List
from copy import copy
from cuqi.density import Density, EvaluatedDensity
from cuqi.distribution import Distribution, Posterior
from cuqi.likelihood import Likelihood
from cuqi.geometry import Geometry, _DefaultGeometry1D
import cuqi
import numpy as np # for splitting array. Can avoid.

class JointDistribution:
    """ 
    Joint distribution of multiple variables.

    Parameters
    ----------
    densities : RandomVariable or Density
        The densities to include in the joint distribution.
        Each density is passed as comma-separated arguments,
        and can be either a :class:'Density' such as :class:'Distribution'
        or :class:`RandomVariable`.

    Notes
    -----
    The joint distribution allows conditioning on any variable of the distribution.
    This is useful for example when Gibbs sampling and the conditionals are needed.
    Conditioning essentially fixes the variable and stores its contribution to the
    log density function.

    Example
    -------

    Consider defining the joint distribution:

    .. math::

        p(y,x,z) = p(y \mid x)p(x \mid z)p(z)

    and conditioning on :math:`y=y_{obs}` leading to the posterior:

    .. math::

        p(x,z \mid y_{obs}) = p(y_{obs} \mid x)p(x \mid z)p(z)

    .. code-block:: python

        import cuqi
        import numpy as np

        y_obs = np.random.randn(10)
        A = np.random.randn(10,3)

        # Define distributions for Bayesian model
        y = cuqi.distribution.Normal(lambda x: A@x, np.ones(10))
        x = cuqi.distribution.Normal(np.zeros(3), lambda z:z)
        z = cuqi.distribution.Gamma(1, 1)

        # Joint distribution p(y,x,z)
        joint = cuqi.distribution.JointDistribution(y, x, z)

        # Posterior p(x,z | y_obs)
        posterior = joint(y=y_obs)
        
    """
    def __init__(self, *densities: [Density, cuqi.experimental.algebra.RandomVariable]):
        """ Create a joint distribution from the given densities. """

        # Check if all RandomVariables are simple (not-transformed)
        for density in densities:
            if isinstance(density, cuqi.experimental.algebra.RandomVariable) and density.is_transformed:
                raise ValueError(f"To be used in {self.__class__.__name__}, all RandomVariables must be untransformed.")

        # Convert potential random variables to their underlying distribution
        densities = [density.distribution if isinstance(density, cuqi.experimental.algebra.RandomVariable) else density for density in densities]

        # Ensure all densities have unique names
        names = [density.name for density in densities]
        if len(names) != len(set(names)):
            raise ValueError("All densities must have unique names.")

        self._densities = list(densities)

        # Make sure every parameter has a distribution (prior)
        cond_vars = self._get_conditioning_variables()
        if len(cond_vars) > 0:
            raise ValueError(f"Every density parameter must have a distribution (prior). Missing prior for {cond_vars}.")

    # --------- Public properties ---------
    @property
    def dim(self) -> List[int]:
        """ Returns the dimensions of the joint distribution. """
        return [dist.dim for dist in self._distributions]

    @property
    def geometry(self) -> List[Geometry]:
        """ Returns the geometries of the joint distribution. """
        return [dist.geometry for dist in self._distributions]

    # --------- Public methods ---------
    def logd(self, *args, **kwargs):
        """ Evaluate the un-normalized log density function. """

        kwargs = self._parse_args_add_to_kwargs(*args, **kwargs)

        # Check that all parameters are passed by matching parameter names with kwargs keys
        if set(self.get_parameter_names()) != set(kwargs.keys()):
            raise ValueError(f"To evaluate the log density function, all parameters must be passed. Received {kwargs.keys()} and expected {self.get_parameter_names()}.")

        # Evaluate the log density function for each density
        logd = 0
        for density in self._densities:
            logd_kwargs = {key:value for (key,value) in kwargs.items() if key in density.get_parameter_names()}
            logd += density.logd(**logd_kwargs)

        return logd

    def __call__(self, *args, **kwargs) -> JointDistribution:
        """ Condition the joint distribution on the given variables. """
        return self._condition(*args, **kwargs)

    def _condition(self, *args, **kwargs): # Public through __call__

        kwargs = self._parse_args_add_to_kwargs(*args, **kwargs)

        # Create new shallow copy of joint density
        new_joint = copy(self) # Shallow copy of self
        new_joint._densities = self._densities[:] # Shallow copy of densities

        # Condition each of the new densities on kwargs relevant to that density
        for i, density in enumerate(new_joint._densities):
            cond_kwargs = {key:value for (key,value) in kwargs.items() if key in density.get_parameter_names()}
            new_joint._densities[i] = density(**cond_kwargs)

        # Potentially reduce joint distribution to a single density
        # This happens if there is only a single parameter left.
        # Can reduce to Posterior, Likelihood or Distribution.
        return new_joint._reduce_to_single_density()

    def get_parameter_names(self) -> List[str]:
        """ Returns the parameter names of the joint distribution. """
        return [dist.name for dist in self._distributions]

    def get_density(self, name) -> Density:
        """ Return a density with the given name. """
        for density in self._densities:
            if density.name == name:
                return density
        raise ValueError(f"No density with name {name}.")

    # --------- Private properties ---------
    @property
    def _distributions(self) -> List[Distribution]:
        """ Returns a list of the distributions (priors) in the joint distribution. """
        return [dist for dist in self._densities if isinstance(dist, Distribution)]

    @property
    def _likelihoods(self) -> List[Likelihood]:
        """ Returns a list of the likelihoods in the joint distribution. """
        return [likelihood for likelihood in self._densities if isinstance(likelihood, Likelihood)]

    @property
    def _evaluated_densities(self) -> List[EvaluatedDensity]:
        """ Returns a list of the evaluated densities in the joint distribution. """
        return [eval_dens for eval_dens in self._densities if isinstance(eval_dens, EvaluatedDensity)]

    # --------- Private methods ---------
    def _get_conditioning_variables(self) -> List[str]:
        """ Return the conditioning variables of the joint distribution. """
        joint_par_names = self.get_parameter_names()
        cond_vars = set()
        for density in self._densities:
            cond_vars.update([par_name for par_name in density.get_parameter_names() if par_name not in joint_par_names])
        return list(cond_vars)

    def _get_fixed_variables(self) -> List[str]:
        """ Return the variables that have been conditioned on (fixed). """
        # Extract names of Likelihoods and EvaluatedDensities
        return [density.name for density in self._densities if isinstance(density, Likelihood) or isinstance(density, EvaluatedDensity)]

    def _parse_args_add_to_kwargs(self, *args, **kwargs):
        """ Parse args and add to kwargs. The args are assumed to follow the order of the parameter names. """
        if len(args)>0:
            ordered_keys = self.get_parameter_names()
            for index, arg in enumerate(args):
                if ordered_keys[index] in kwargs:
                    raise ValueError(f"{ordered_keys[index]} passed as both argument and keyword argument.\nArguments follow the listed parameter names order: {ordered_keys}")
                kwargs[ordered_keys[index]] = arg
        return kwargs

    def _sum_evaluated_densities(self):
        """ Return the sum of the evaluated densities in the joint distribution """
        return sum([density.logd() for density in self._evaluated_densities])

    def _reduce_to_single_density(self):
        """ Reduce the joint distribution to a single density if possible.

        The single density is either a Posterior, Likelihood or Distribution.
        
        This method is a hack to allow our current samplers to work with
        the joint distribution. It should be removed in the future.
        """
        # Count number of distributions and likelihoods
        n_dist = len(self._distributions)
        n_likelihood = len(self._likelihoods)

        # Cant reduce if there are multiple distributions or likelihoods
        if n_dist > 1:
            return self

        # If exactly one distribution and multiple likelihoods reduce
        if n_dist == 1 and n_likelihood > 1:
            return MultipleLikelihoodPosterior(*self._densities)
        
        # If exactly one distribution and one likelihood its a Posterior
        if n_dist == 1 and n_likelihood == 1:
            # Ensure parameter names match, otherwise return the joint distribution
            if set(self._likelihoods[0].get_parameter_names()) != set(self._distributions[0].get_parameter_names()):
                return self
            return self._add_constants_to_density(Posterior(self._likelihoods[0], self._distributions[0]))

        # If exactly one distribution and no likelihoods its a Distribution
        if n_dist == 1 and n_likelihood == 0:
            return self._add_constants_to_density(self._distributions[0])        
        
        # If no distributions and exactly one likelihood its a Likelihood
        if n_likelihood == 1 and n_dist == 0:
            return self._likelihoods[0]

        # If only evaluated densities left return joint to ensure logd method is available
        if n_dist == 0 and n_likelihood == 0:
            return self
        
    def _add_constants_to_density(self, density: Density):
        """ Add the constants (evaluated densities) to a single density. Used when reducing to single density. """

        if isinstance(density, EvaluatedDensity):
            raise ValueError("Cannot add the sum of all evaluated densities to an EvaluatedDensity.")

        density._constant += self._sum_evaluated_densities()
        return density

    def _as_stacked(self) -> _StackedJointDistribution:
        """ Return a stacked JointDistribution with the same densities. """
        return _StackedJointDistribution(*self._densities)

    def __repr__(self):
        msg = f"JointDistribution(\n"
        msg += "    Equation: \n\t"

        # Construct equation expression
        joint_par_names = ",".join(self.get_parameter_names())
        fixed_par_names = ",".join(self._get_fixed_variables())
        if len(joint_par_names) == 0:
            msg += "Constant number"
        else:
            # LHS of equation: p(x,y,z) = or p(x,y|z) ∝
            msg += f"p({joint_par_names}"
            if len(fixed_par_names) > 0:
                msg += f"|{fixed_par_names}) ∝ "
            else:
                msg += ") = "

            # RHS of equation: product of densities
            for density in self._densities:
                par_names = ",".join([density.name])
                cond_vars = ",".join(set(density.get_parameter_names())-set([density.name]))

                # Distributions are written as p(x|y) or p(x). Likelihoods are L(y|x).
                # x=par_names, y=cond_vars.
                if isinstance(density, Likelihood):
                    msg += f"L({cond_vars}|{par_names})"
                elif isinstance(density, Distribution):
                    msg += f"p({par_names}"
                    if len(cond_vars) > 0:
                        msg += f"|{cond_vars}"
                    msg += ")"
        
        msg += "\n"
        msg += "    Densities: \n"

        # Create "Bayesian model" equations
        for density in self._densities:
            msg += f"\t{density.name} ~ {density}\n"

        # Wrap up
        msg += " )"

        return msg


class _StackedJointDistribution(JointDistribution, Distribution):
    """ A joint distribution where all parameters are stacked into a single vector.

    This acts like a regular Distribution with a single parameter vector even
    though it is actually a joint distribution.

    This is intended to be used by samplers that are not able to handle
    joint distributions. A joint distribution can be converted to a stacked
    joint distribution by calling the :meth:`_as_stacked` method.
    
    See :class:`JointDistribution` for more details on the joint distribution.
    """

    @property
    def dim(self):
        return sum(super().dim)

    @property
    def geometry(self):
        return _DefaultGeometry1D(self.dim)

    def logd(self, stacked_input):
        """ Return the un-normalized log density function stacked joint density. """

        # Split the stacked input into individual inputs and call superclass
        split_indices = np.cumsum(super().dim)  # list(accumulate(super().dim))
        inputs = np.split(stacked_input, split_indices[:-1])
        names = self.get_parameter_names()

        # Create keyword arguments
        kwargs = dict(zip(names, inputs))

        return super().logd(**kwargs)

    def logpdf(self, stacked_input):
        return self.logd(stacked_input)
    
    def _sample(self, Ns=1):
        raise TypeError(f"{self.__class__.__name__} does not support sampling.")

    def __repr__(self):
        return "_Stacked"+super().__repr__()


class MultipleLikelihoodPosterior(JointDistribution, Distribution):
    """ A posterior distribution with multiple likelihoods and a single prior.

    Parameters
    ----------
    densities : :class:`Distribution` or :class:`~cuqi.likelihood.Likelihood`
        The densities that make up the Posterior. Must include
        at least three densities. For a simple Likelihood and prior
        use :class:`Posterior` instead.

    Notes
    -----    
    This acts like a regular distribution with a single parameter vector. Behind-the-scenes
    it is a joint posterior distribution with multiple likelihoods and a single prior.
    This is mostly intended to be used by samplers that are not able to handle joint distributions. 
    See :class:`JointDistribution` for more details on the joint distribution.   
    
    """

    def __init__(self, *densities: Density):
        super().__init__(*densities)
        self._check_densities_have_same_parameter()

    @property
    def geometry(self):
        """ The geometry of the distribution. """
        return self.prior.geometry

    @property
    def dim(self):
        """ Return the dimension of the distribution. """
        return self.prior.dim

    @property
    def prior(self):
        """ Return the prior distribution of the posterior. """
        return self._distributions[0]

    @property
    def likelihoods(self):
        """ Return the likelihoods of the posterior. """
        return self._likelihoods

    @property
    def models(self):
        """ Return the forward models that make up the posterior. """
        return [likelihood.model for likelihood in self.likelihoods]

    def logpdf(self, *args, **kwargs):
        return self.logd(*args, **kwargs)

    def gradient(self, *args, **kwargs):
        """ Return the gradient of the un-normalized log density function. """
        return sum(density.gradient(*args, **kwargs) for density in self._densities)      

    def _sample(self, Ns=1):
        raise TypeError(f"{self.__class__.__name__} does not support direct sampling.")

    def _check_densities_have_same_parameter(self):
        """ Checks the densities if they are for one parameter only and that there are at least 3 densities. """

        if len(self._densities) < 3:
            raise ValueError(f"{self.__class__.__name__} requires at least three densities. For a single likelihood and prior use Posterior instead.")
        
        if len(self.likelihoods) == 0:
            raise ValueError(f"{self.__class__.__name__} must have a likelihood and prior.")

        # Check that there is only a single parameter
        par_names = self.get_parameter_names()
        if len(set(par_names)) > 1:
            raise ValueError(f"{self.__class__.__name__} requires all densities to have the same parameter name.")

    def __repr__(self):
        # Remove first line of super repr and add class name to the start
        return f"{self.__class__.__name__}(\n" + "\n".join(super().__repr__().split("\n")[1:])



import cuqi
import inspect
import numpy as np

# This import makes suggest_sampler easier to read
import cuqi.experimental.mcmc as samplers 


class SamplerRecommender(object):
    """
    This class can be used to automatically choose a sampler.

    Parameters
    ----------
    target: Density or JointDistribution
        Distribution to get sampler recommendations for.

    exceptions: list[cuqi.experimental.mcmc.Sampler], *optional*
        Samplers not to be recommended.
    
    Example
    -------
    .. code-block:: python
        import numpy as np
        from cuqi.distribution import Gamma, Gaussian, JointDistribution
        from cuqi.experimental import SamplerRecommender

        x = Gamma(1, 1)
        y = Gaussian(np.zeros(2), cov=lambda x: 1 / x)
        target = JointDistribution(y, x)(y=1)

        recommender = SamplerRecommender(target)
        valid_samplers = recommender.valid_samplers()
        recommended_sampler = recommender.recommend()
        print("Valid samplers:", valid_samplers)
        print("Recommended sampler:\n", recommended_sampler)

    """

    def __init__(self, target:cuqi.density.Density, exceptions = []):
        self._target = target
        self._exceptions = exceptions
        self._create_ordering()

    @property
    def target(self) -> cuqi.density.Density:
        """ Return the target Distribution. """
        return self._target

    @target.setter
    def target(self, value:cuqi.density.Density):
        """ Set the target Distribution. Runs validation of the target. """
        if value is None:
            raise ValueError("Target needs to be of type cuqi.density.Density.")
        self._target = value

    def _create_ordering(self):
        """
        Every element in the ordering consists of a tuple:
        (
            Sampler: Class
            boolean: additional conditions on the target
            parameters: additional parameters to be passed to the sampler once initialized
        )
        """
        number_of_components = np.sum(self._target.dim)

        self._ordering = [
            # Direct and Conjugate samplers
            (samplers.Direct, True, {}),
            (samplers.Conjugate, True, {}),
            (samplers.ConjugateApprox, True, {}),
            # Specialized samplers
            (samplers.LinearRTO, True, {}),
            (samplers.RegularizedLinearRTO, True, {}),
            (samplers.UGLA, True, {}),
            # Gradient.based samplers (Hamiltonian and Langevin)
            (samplers.NUTS, True, {}),
            (samplers.MALA, True, {}),
            (samplers.ULA, True, {}),
            # Gibbs and Componentwise samplers
            (samplers.HybridGibbs, True, {"sampling_strategy" : self.recommend_HybridGibbs_sampling_strategy(as_string = False)}),
            (samplers.CWMH, number_of_components <= 100, {"scale" : 0.05*np.ones(number_of_components),
                            "initial_point" : 0.5*np.ones(number_of_components)}),
            # Proposal based samplers
            (samplers.PCN, True, {"scale" : 0.02}),
            (samplers.MH, number_of_components <= 1000, {}),
        ]

    @property
    def ordering(self):
        """ Returns the ordered list of recommendation rules used by the recommender. """
        return self._ordering

    def valid_samplers(self, as_string = True):
        """
        Finds all possible samplers that can be used for sampling from the target distribution.

        Parameters
        ----------

        as_string : boolean
            Whether to return the name of the sampler as a string instead of instantiating a sampler. *Optional*

        """

        all_samplers = [(name, cls) for name, cls in inspect.getmembers(cuqi.experimental.mcmc, inspect.isclass) if issubclass(cls, cuqi.experimental.mcmc.Sampler)]
        valid_samplers = []

        for name, sampler in all_samplers:
            try:
                sampler(self.target)
                valid_samplers += [name if as_string else sampler]
            except:
                pass

        # Need a separate case for HybridGibbs
        if self.valid_HybridGibbs_sampling_strategy() is not None:
            valid_samplers += [cuqi.experimental.mcmc.HybridGibbs.__name__ if as_string else cuqi.experimental.mcmc.HybridGibbs]

        return valid_samplers

    def valid_HybridGibbs_sampling_strategy(self, as_string = True):
        """
            Find all possible sampling strategies to be used with the HybridGibbs sampler.
            Returns None if no sampler could be suggested for at least one conditional distribution.

            Parameters
            ----------

            as_string : boolean
                Whether to return the name of the samplers in the sampling strategy as a string instead of instantiating samplers. *Optional*
        

        """

        if not isinstance(self.target, cuqi.distribution.JointDistribution):
            return None

        par_names = self.target.get_parameter_names()

        valid_samplers = dict()
        for par_name in par_names:
            conditional_params = {par_name_: np.ones(self.target.dim[i]) for i, par_name_ in enumerate(par_names) if par_name_ != par_name}
            conditional = self.target(**conditional_params)

            recommender = SamplerRecommender(conditional)
            samplers = recommender.valid_samplers(as_string)
            if len(samplers) == 0:
                return None

            valid_samplers[par_name] = samplers

        return valid_samplers

    def recommend(self, as_string = False):
        """
            Suggests a possible sampler that can be used for sampling from the target distribution.
            Return None if no sampler could be suggested.

            Parameters
            ----------

            as_string : boolean
                Whether to return the name of the sampler as a string instead of instantiating a sampler. *Optional*

        """

        valid_samplers = self.valid_samplers(as_string = False)

        for suggestion, flag, values in self._ordering:
            if flag and (suggestion in valid_samplers) and (suggestion not in self._exceptions):
                # Sampler found
                if as_string:
                    return suggestion.__name__
                else:
                    return suggestion(self.target, **values)

        # No sampler can be suggested
        raise ValueError("Cannot suggest any sampler. Either the provided distribution is incorrectly defined or there are too many exceptions provided.")

    def recommend_HybridGibbs_sampling_strategy(self, as_string = False):
        """
            Suggests a possible sampling strategy to be used with the HybridGibbs sampler.
            Returns None if no sampler could be suggested for at least one conditional distribution.

            Parameters
            ----------

            target : `cuqi.distribution.JointDistribution`
                The target distribution get a sampling strategy for.
            
            as_string : boolean
                Whether to return the name of the samplers in the sampling strategy as a string instead of instantiating samplers. *Optional*
        
        """

        if not isinstance(self.target, cuqi.distribution.JointDistribution):
            return None

        par_names = self.target.get_parameter_names()

        suggested_samplers = dict()
        for par_name in par_names:
            conditional_params = {par_name_: np.ones(self.target.dim[i]) for i, par_name_ in enumerate(par_names) if par_name_ != par_name}
            conditional = self.target(**conditional_params)

            recommender = SamplerRecommender(conditional, exceptions = self._exceptions.copy())
            sampler = recommender.recommend(as_string = as_string)

            if sampler is None:
                return None

            suggested_samplers[par_name] = sampler

        return suggested_samplers



from cuqi.array import CUQIarray
from cuqi.density import Density
import numpy as np
import inspect
from numbers import Number
from scipy.sparse import issparse, diags
from scipy.sparse import linalg as spslinalg
from dataclasses import dataclass
from abc import ABCMeta
import copy
import matplotlib.pyplot as plt


def force_ndarray(value,flatten=False):
    if not isinstance(value, np.ndarray) and value is not None and not issparse(value) and not callable(value):
        if hasattr(value,'__len__') and len(value)>1:
            value = np.array(value)
        else:
            value = np.array(value).reshape((1,1))
            
        if flatten is True:
            value = value.flatten()
    if isinstance(value,np.matrix): #Convert to array if matrix (matrix acts different on (n,) arrays)
        value = value.A
    return value

def infer_len(value):
    """ Infer the length of the given input value.

    Matrices are assumed to have length equal to the number of rows.
    Numbers are considered to have length 1.
    Other objects with no length are considered to have length 0.
    """
    if hasattr(value,'__len__'):
        try:
            return len(value)
        except TypeError: #Special-case for scipy sparse matrices, which have len but return an error
            return value.shape[0]
    elif isinstance(value, Number):
        return 1
    else:
        return 0

def get_non_default_args(func):
    """ Returns the non-default arguments and kwargs from a callable function"""
    # If the function has variable _non_default_args, use that for speed.
    if hasattr(func, '_non_default_args'):
        return func._non_default_args

    # Otherwise, get the arguments from the function signature.
    sig = inspect.signature(func)
    para = sig.parameters

    nonDefaultArgs = []
    for key in para:
        if key != "kwargs" and key != "args" and para[key].default is inspect._empty: #no default and not kwargs
            nonDefaultArgs.append(key)
    return nonDefaultArgs


def get_direct_attributes(dist):
    keys = vars(dist).keys()
    return [key for key in keys]

def get_indirect_variables(dist):
    attributes = []
    for attribute in dist.get_mutable_variables():
        value = getattr(dist, attribute)
        if check_if_conditional_from_attr(value):
            keys = get_non_default_args(value)
            for key in keys:
                if key not in attributes: #Ensure we did not already find this key
                    attributes.append(key)
    return attributes 

def check_if_conditional_from_attr(value):
    """
    Check if a distribution is conditional from a given attribute.
    So far, we assume that a distribution is conditional if
    - the given attribute is a callable function and
    - the given attribute is not a LinearOperator.
    """
    if isinstance(value, spslinalg.LinearOperator):
        return False
    elif callable(value):
        return True
    else:
        return False

def get_writeable_attributes(dist):
    """ Get writeable attributes of object instance. """
    attributes = []
    for key in vars(dist).keys():
        if key[0] != "_":
            attributes.append(key)
    return attributes

def get_writeable_properties(cls, stop_at_class=object):
    """ Get writeable properties of class type."""

    # Potentially convert object instance to class type.
    if isinstance(cls, stop_at_class) and isinstance(type(cls), ABCMeta):
        cls = type(cls)

    # Compute writeable properties of this class
    writeable_properties = [attr for attr, value in vars(cls).items()
                 if isinstance(value, property) and value.fset is not None]

    # Stop recursion at stop_at_class
    if cls == stop_at_class:
        return writeable_properties

    # Recursively get writeable properties of parents
    for base in cls.__bases__:
        writeable_properties += get_writeable_properties(base)
    return writeable_properties

@dataclass
class ProblemInfo:
    """Problem info dataclass. Gives a convenient way to store data defined in test-problems."""
    exactSolution: np.ndarray = None
    exactData: np.ndarray = None
    Miscellaneous: dict = None
    infoString: str = None

    def __repr__(self) -> str:
        out_str = "ProblemInfo with the following set attributes:\n"+str(self.getSetAttributes())      
        if self.infoString is not None:
            out_str = out_str+"\n infoString: "+str(self.infoString)
        if self.Miscellaneous is not None:
            out_str = out_str+f"\n Miscellaneous: {self.Miscellaneous.keys()}"
        return out_str

    def getSetAttributes(self):
        """Returns a list of all attributes that are not None."""
        dict = vars(self)
        return list({key for key in dict if dict[key] is not None})

# work-around to compute sparse Cholesky
def sparse_cholesky(A):
    """Computes Cholesky factorization for sparse matrix `A` and returns the upper triangular factor `U`, where `A=U^T@U`"""
    # https://gist.github.com/omitakahiro/c49e5168d04438c5b20c921b928f1f5d
    LU = spslinalg.splu(A, diag_pivot_thresh=0, permc_spec='natural') # sparse LU decomposition


    # check the matrix A is positive definite
    if (LU.perm_r == np.arange(A.shape[0])).all() and (LU.U.diagonal() > 0).all(): 
        return (LU.L @ (diags(LU.U.diagonal()**0.5))).T
    else:
        raise TypeError('The matrix is not positive semi-definite')

def approx_derivative(func, wrt, direction=None, epsilon=np.sqrt(np.finfo(float).eps)):
    """Approximates the derivative of callable (possibly vector-valued) function `func` evaluated at point `wrt`. If `direction` is provided, the direction-Jacobian product will be computed and returned, otherwise, the Jacobian matrix (or the gradient in case of a scalar function `func`) will be returned. The approximation is done using forward differences.

    Parameters
    ----------
    func: function handler
        A vector-valued function of the form func(x).

    wrt : ndarray
        The point at which the derivative to be evaluated.

    direction : ndarray
        The direction used to compute direction-Jacobian product. 
        If None, the Jacobian matrix is returned.

    epsilon: float
        The spacing in the finite difference approximation.

    Returns
    -------
    ndarray
        The approximate Jacobian matrix.
    """
    # Raise an error if wrt or direction is a CUQIarray.
    # Example of scenario where this is needed: 
    # the line Matr[i] = (func(wrt+dx) - f0)/epsilon
    # does not give correct results if for example
    # wrt is a CUQIarray with is_par=False and its 
    # corresponding geometry par2fun map is not identity
    # (e.g. funvalues=paramters**2), because wrt entries 
    # are interpreted as function value.

    if isinstance(wrt, CUQIarray) or isinstance(direction, CUQIarray):
        raise NotImplementedError("approx_derivative is not implemented"+
                                   "for inputs of type CUQIarray")

    # We compute the Jacobian matrix of func using forward differences.
    # If the function is scalar-valued, we compute the gradient instead.
    # If the direction is provided, we compute the direction-Jacobian product.
    wrt = force_ndarray(wrt, flatten=True)
    f0 = func(wrt)
    Matr = np.zeros([infer_len(wrt), infer_len(f0)])
    dx = np.zeros(len(wrt))

    # Compute the Jacobian matrix (transpose)
    for i in range(len(wrt)):
        dx[i] = epsilon
        Matr[i] = (func(wrt+dx) - f0)/epsilon
        dx[i] = 0.0

    # Return the Jacobian matrix (or the gradient)
    # or the direction-Jacobian product
    if direction is None:
        if infer_len(f0) == 1:
            return Matr.reshape(infer_len(wrt))
        else:
            return Matr.T
    else:
        return Matr@direction

def approx_gradient(func, x, epsilon= 0.000001):
    """Approximates the gradient of callable scalar function `func` evaluated at point `x`. The approximation is done using finite differences with
    step size `epsilon`."""
    
    # Derivative of a scalar function using forward differences
    if isinstance(x, Number):
        return (func(x+epsilon) - func(x))/epsilon

    # Initialize variables
    FD_gradient = x*0.0
    eps_vec = x*0.0
    func_x = func(x)

    # Compute the gradient using forward differences component by component
    x_len = infer_len(x)
    for i in range(x_len):
        eps_vec[i] = epsilon
        x_plus_eps = x + eps_vec
        FD_gradient[i] = (func(x_plus_eps) - func_x)/epsilon
        eps_vec[i] = 0.0
        
    return FD_gradient


# Function for plotting 1D density functions
def plot_1D_density(density:Density,
                    v_min, v_max,
                    N=501, log_scale=False,
                    **kwargs):
    """ Plot 1D density function 

    Parameters
    ----------
    density : CUQIpy Density
        The density to be plotted.

    v_min : float
        Minimum value for the variable.
    
    v_max : float
        Maximum value for the variable.

    N : int
        Number of grid points for the variable.

    log_scale : bool
        If True, the density is plotted in log scale.
    
    kwargs : dict
        Additional keyword arguments for the plot that are passed to the
        underlying plotting method: `matplotlib.pyplot.plot` function 
        in this case.

    """
    # Assert that the density is 1D
    assert density.dim == 1, "The density must be for a scalar variable"
    ls = np.linspace(v_min, v_max, N)

    # Create a map to evaluate density
    density_map = (lambda x: x) if log_scale else (lambda x: np.exp(x))

    # Evaluate density on grid
    y = [density_map(density.logd(grid_point)) for grid_point in ls]
    p = plt.plot(ls, y, **kwargs)
    return p

# Function for plotting 2D density functions
def plot_2D_density(density: Density, 
                    v1_min, v1_max,
                    v2_min, v2_max,
                    N1=201, N2=201,
                    log_scale=False,
                    **kwargs):
    """ Plot 2D density function 

    Parameters
    ----------
    density : CUQIpy Density
        The density to be plotted.

    v1_min : float
        Minimum value for the first variable.
    
    v1_max : float
        Maximum value for the first variable.
    
    v2_min : float
        Minimum value for the second variable.

    v2_max : float  
        Maximum value for the second variable.

    N1 : int
        Number of grid points for the first variable.

    N2 : int
        Number of grid points for the second variable.
    
    log_scale : bool
        If True, the density is plotted in log scale.

    kwargs : dict
        Additional keyword arguments for the plot that are passed to the
        underlying plotting method: `matplotlib.pyplot.imshow` function 
        in this case.

    """
    # Assert that the density is 2D
    assert density.dim == 2,\
        "The density must be for a two-dimensional variable"
    # Create grid
    ls1 = np.linspace(v1_min, v1_max, N1)
    ls2 = np.linspace(v2_min, v2_max, N2)
    grid1, grid2 = np.meshgrid(ls1, ls2)

    # Create a map to evaluate density
    density_map = (lambda x: x) if log_scale else (lambda x: np.exp(x))

    # Evaluate density on grid
    evaluated_density = np.zeros((N1, N2))
    for ii in range(N1):
        for jj in range(N2):
            evaluated_density[ii,jj] = density_map(
                density.logd([grid1[ii,jj], grid2[ii,jj]])) 

    # Plot
    pixelwidth_x = (v1_max-v1_min)/(N2-1)
    pixelwidth_y = (v2_max-v2_min)/(N2-1)

    hp_x = 0.5*pixelwidth_x
    hp_y = 0.5*pixelwidth_y

    extent = (v1_min-hp_x, v1_max+hp_x, v2_min-hp_y, v2_max+hp_y)

    im = plt.imshow(evaluated_density, origin='lower', extent=extent, **kwargs)
    return im



def count_nonzero(x, threshold = 1e-6):
        """ Returns the number of values in an array whose absolute value is larger than a specified threshold

        Parameters
        ----------
        x : `np.ndarray` 
            Array to count nonzero elements of.

        threshold : float
            Theshold for considering a value as nonzero.
        """
        return np.sum([np.abs(v) >= threshold for v in x])

        
def count_within_bounds(x, lower_bounds, upper_bounds, threshold = 1e-6, exception = np.nan):
        """ Returns the number of values in an array whose value lies between the provided lower and upper bounds.

        Parameters
        ----------
        x : `np.ndarray` 
            Array to count elements of.

        lower_bounds : `np.ndarray` 
            Lower bound on values to disregard when counting.

        upper_bounds : `np.ndarray` 
            Upper bound on values to disregard.

        threshold : float
            Theshold for considering a value as nonzero.
        """
        return np.sum([l + threshold <= v and v <= u - threshold and not (exception - threshold <= v and v <= exception + threshold) for v, l, u in zip(x, lower_bounds, upper_bounds)])
        
    
def count_constant_components_1D(x, threshold = 1e-2, lower = -np.inf, upper = np.inf, exception = np.nan):
        """ Returns the number of piecewise constant components in a one-dimensional array

        Parameters
        ----------
        x : `np.ndarray` 
            1D Array to count components of.

        threshold : float
            Strict theshold on when the difference of neighbouring values is considered zero.

        lower : float or numpy.ndarray
            Piecewise constant components below this value are not counted.

        upper : float or numpy.ndarray
            Piecewise constant components above this value are not counted.
        """

        if not isinstance(lower, np.ndarray):
            lower = lower*np.ones_like(x)

        if not isinstance(upper, np.ndarray):
            upper = upper*np.ones_like(x)

        counter = 0
        index = 0
        if (x[index] > lower[index] and
            x[index] < upper[index] and
            not (exception - threshold <= x[index] and x[index] <= exception + threshold)):
            counter += 1
        
        x_previous = x[0]

        for x_current in x[1:]:
            index += 1
            if (abs(x_previous - x_current) >= threshold and
                x_current > lower[index] and
                x_current < upper[index] and
                not (exception - threshold <= x_current and x_current <= exception + threshold)):
                    counter += 1

            x_previous = x_current
    
        return counter

def count_constant_components_2D(x, threshold = 1e-2, lower = -np.inf, upper = np.inf):
        """ Returns the number of piecewise constant components in a two-dimensional array

        Parameters
        ----------
        x : `np.ndarray` 
            2D Array to count components of.

        threshold : float
            Strict theshold on when the difference of neighbouring values is considered zero.

        lower : float or numpy.ndarray
            Piecewise constant components below this value are not counted.

        upper : float or numpy.ndarray
            Piecewise constant components above this value are not counted.
        """

        if not isinstance(lower, np.ndarray):
            lower = lower*np.ones_like(x)

        if not isinstance(upper, np.ndarray):
            upper = upper*np.ones_like(x)

        filled = np.zeros_like(x, dtype = int)
        counter = 0

        def process(i, j):
            queue = []
            queue.append((i,j))
            filled[i, j] = 1
            while len(queue) != 0:
                (icur, jcur) = queue.pop(0)
                
                if icur > 0 and filled[icur - 1, jcur] == 0 and abs(x[icur, jcur] - x[icur - 1, jcur]) <= threshold:
                    filled[icur - 1, jcur] = 1
                    queue.append((icur-1, jcur))
                if jcur > 0 and filled[icur, jcur-1] == 0 and abs(x[icur, jcur] - x[icur, jcur - 1]) <= threshold:
                    filled[icur, jcur-1] = 1
                    queue.append((icur, jcur-1))
                if icur < x.shape[0]-1 and filled[icur + 1, jcur] == 0 and abs(x[icur, jcur] - x[icur + 1, jcur]) <= threshold:
                    filled[icur + 1, jcur] = 1
                    queue.append((icur+1, jcur))
                if jcur < x.shape[1]-1 and filled[icur, jcur + 1] == 0 and abs(x[icur, jcur] - x[icur, jcur + 1]) <= threshold:
                    filled[icur, jcur + 1] = 1
                    queue.append((icur, jcur+1))
        
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                if filled[i,j] == 0:
                    if x[i,j] > lower[i,j] and x[i,j] < upper[i,j]:
                        counter += 1
                    process(i, j)
        return counter
                    


def piecewise_linear_1D_DoF(x, threshold = 1e-5, exception_zero = False, exception_flat = False):
        """ Returns the degrees of freedom of a piecewise linear signal.
        Assuming linear interpolation, this corresponds to the number of non-differentiable points, including end-points.

        Parameters
        ----------
        x : `np.ndarray` 
            1D Array to compute degrees of freedom of.

        threshold : float
            Strict theshold on when values are considered zero.

        exception_zero : Boolean
            Whether a zero piecewise linear components should be considered.

        exception_flat : Boolean
            Whether a flat piecewise linear components should be considered.
        """

        differences = x[1:] - x[:-1]
        double_differences = differences[1:] - differences[:-1]

        joints = [True] + [np.abs(d) >= threshold for d in double_differences] + [True]
        
        if not exception_zero and not exception_flat:
            return np.sum(joints)
        elif exception_zero:
            return np.sum(joints and [np.abs(v) < threshold for v in x])
        elif exception_flat:
            prev_joint = None
            counter = 0
            for i in range(len(joints)):
                if joints[i]:
                    counter += 1
                    if prev_joint is None:
                        prev_joint = i
                    else:
                        if np.abs(x[i] - x[prev_joint]) < threshold:
                            counter -= 1
                        prev_joint = i
            return counter



import cuqi
import pytest
import numpy as np

def test_BayesianProblem_init():
    """ Test BayesianProblem initialization options for likelihood + prior """
    
    f = lambda x: np.hstack((x+5, x-5))
    A = cuqi.model.Model(f, 10, 5)
    x = cuqi.distribution.Gaussian(np.zeros(A.domain_dim), 1)
    y = cuqi.distribution.Gaussian(A, 1)
    y_obs = y(x.sample()).sample() # Sample prior, then sample data distribution

    # Posterior manually
    P = cuqi.distribution.Posterior(y.to_likelihood(y_obs), x)

    # Various ways to define BP
    BP1 = cuqi.problem.BayesianProblem(x, y).set_data(y=y_obs)    # Set data method
    BP2 = cuqi.problem.BayesianProblem(x, y, y=y_obs)             # data as keyword
    BP3 = cuqi.problem.BayesianProblem(y.to_likelihood(y_obs), x) # likelihood, prior

    # Check that all are the same
    _assert_same_BP_posterior(BP1, P)    
    _assert_same_BP_posterior(BP2, P)
    _assert_same_BP_posterior(BP3, P)

def _assert_same_BP_posterior(BP: cuqi.problem.BayesianProblem, P: cuqi.distribution.Posterior):
    assert np.allclose(BP.posterior.prior.mean, P.prior.mean)
    assert np.allclose(BP.posterior.prior.cov, P.prior.cov)
    assert BP.posterior.likelihood.distribution.mean == P.likelihood.distribution.mean # model
    assert np.allclose(BP.posterior.likelihood.distribution.cov, P.likelihood.distribution.cov)

def test_BayesianProblem_wrong_names():
    """ Test that wrong names raise errors """

    z = cuqi.distribution.Gaussian(np.zeros(10), 1) # p(z)
    y = cuqi.distribution.Gaussian(lambda x: x, 1)  # p(y|x)

    with pytest.raises(ValueError, match=r"Missing prior for"):
        cuqi.problem.BayesianProblem(z, y)

def test_BayesianProblem_hier_errors():
    """ Test that hierarchical Bayes models raise various errors (for now) """

    d = cuqi.distribution.Gamma(1, 1)
    x = cuqi.distribution.Gaussian(np.zeros(10), lambda d: d)
    y = cuqi.distribution.Gaussian(lambda x: x, 1)

    BP = cuqi.problem.BayesianProblem(y, x, d)

    with pytest.raises(ValueError, match=r"Unable to extract posterior"):
        BP.posterior

    with pytest.raises(ValueError, match=r"Unable to extract likelihood"):
        BP.likelihood

    with pytest.raises(ValueError, match=r"Unable to extract prior"):
        BP.prior

    with pytest.raises(ValueError, match=r"Unable to extract prior"):
        BP.sample_prior(10)

    with pytest.raises(NotImplementedError, match=r"Unable to determine sampling strategy"):
        BP.UQ()

    with pytest.raises(NotImplementedError, match=r"Unable to determine sampling strategy"):
        BP.sample_posterior(10)

    with pytest.raises(ValueError, match=r"Unable to extract prior"): # Should rly be posterior
        BP.MAP()

    with pytest.raises(ValueError, match=r"Unable to extract likelihood"):
        BP.ML()

    with pytest.raises(ValueError, match=r"Unable to extract likelihood"):
        BP.data

    with pytest.raises(ValueError, match=r"Unable to extract likelihood"):
        BP.model

def test_cannot_set_data():

    x = cuqi.distribution.Gaussian(0, 1)
    y = cuqi.distribution.Gaussian(lambda x: x, 1)

    BP = cuqi.problem.BayesianProblem(y, x)

    # Joint distribution does not care if we set the wrong data name (nothing happens)
    # We add a check here to catch this in the future when its fixed.
    BP.set_data(z=1) # z is not a valid name, should raise ValueError in future

    # Set data
    BP.set_data(y=1)

    # Try to set data again
    with pytest.raises(ValueError, match=r"data is already set"):
        BP.set_data(y=1)

def test_BayesianProblem_geometry_consistency():

    domain_geometry = cuqi.geometry.StepExpansion(np.arange(5), 3)
    range_geometry = cuqi.geometry.Continuous1D(10)

    f = lambda x: np.hstack((x+5, x-5))
    A = cuqi.model.Model(f, range_geometry, domain_geometry)
    x = cuqi.distribution.Gaussian(np.zeros(A.domain_dim), 1)
    y = cuqi.distribution.Gaussian(A, 1)
    y_obs = y(x.sample()).sample() # Sample prior, then sample data distribution

    # BayesianProblem
    BP = cuqi.problem.BayesianProblem(x, y).set_data(y=y_obs)

    # Check that we do not change the user provided geometries (We may want to change these in BP!)
    # This is what the now removed _check_geometry_consistency did (but it did not do it for all cases)
    # The tasks should be delegated to Posterior or JointDistribution
    assert BP.posterior.prior.geometry == x.geometry # prior geometry is not changed
    assert BP.posterior.likelihood.distribution.geometry == y.geometry # likelihood geometry is not changed
    assert BP.data.geometry == y_obs.geometry # data geometry is not changed
    assert BP.model == A # model geometry is not changed
    
    # Check that BayesianProblem correctly infers the geometry of the posterior (from model)
    assert BP.posterior.geometry == domain_geometry
    assert BP.sample_posterior(10).geometry == domain_geometry
    #assert BP.sample_prior(10).geometry == domain_geometry # This fails because prior is sampled directly. TODO (should be domain_geometry)
    assert BP.MAP().geometry == domain_geometry
    assert BP.ML().geometry == domain_geometry


def test_passing_burnin_to_UQ_method(capfd):
    """ Test that passing burnin to UQ method works correctly"""
    # Create simple forward model
    fwd_model = cuqi.model.Model(lambda x: x, 1, 1)

    # Create prior and data distributions
    x = cuqi.distribution.Gaussian(0, 1)
    y = cuqi.distribution.Gaussian(fwd_model(x), 1)

    # Create BayesianProblem and set data
    BP = cuqi.problem.BayesianProblem(x, y)
    BP.set_data(y=1)

    # Apply UQ method with burnin
    samples = BP.UQ(Ns=10, Nb=5) 

    # Read output
    out, err = capfd.readouterr()

    # Check that correct burnin is used
    assert "50%" in out
    assert "Sample 15 / 15" in out



import pytest
from scipy.linalg import cholesky
from scipy.sparse import diags
from cuqi.utilities import sparse_cholesky, plot_1D_density, plot_2D_density, count_nonzero, count_within_bounds, count_constant_components_1D, count_constant_components_2D, piecewise_linear_1D_DoF
from cuqi.model import LinearModel
from cuqi.distribution import Gaussian, Uniform, JointDistribution
import numpy as np
import matplotlib.pyplot as plt


@pytest.mark.parametrize("P", [
    diags([-1, 2, -1], [-1, 0, 1], shape=(128, 128)),
    diags([1, -4, 6, -4, 1], [-2, -1, 0, 1, 2], shape=(128, 128))
])
def test_sparse_cholesky(P):
    """ Test the sparse Cholesky decomposition. P is a sparse matrix (often precision). """
    # Scipy version (on dense only)
    L1 = cholesky(P.toarray())

    # Scipy-based version from CUQIpy (on sparse)
    L2 = sparse_cholesky(P) 

    assert np.allclose(L1, L2.toarray()) # Convert to dense to compare


def posterior_2D_components():
    """A function to create and return 2D Bayesian posterior components"""
    # Create a CUQIpy model
    A = LinearModel(np.array([[1.0, 1.0]]))
    # Create a CUQIpy prior
    x = Gaussian(mean=np.array([0.0, 0.0]), cov=np.array([1.0, 1.0]))
    # Create a CUQIpy likelihood
    y = Gaussian(A@x, cov=0.1)
    return x, y


def likelihood_2D():
    """A function to create and return a 2D likelihood"""
    x, y = posterior_2D_components()
    likelihood = y.to_likelihood(2)
    return likelihood


def posterior_2D():
    """A function to create and return a 2D posterior"""
    x, y = posterior_2D_components()
    joint = JointDistribution(x, y)
    posterior = joint(y=2)
    return posterior


@pytest.mark.parametrize("density", [
    Gaussian(mean=0.0, cov=1.0),
    Uniform(low=0.0, high=1.0),
    Gaussian(mean=np.array([0.0, 0.0]), cov=np.array([1.0, 1.0])),
    likelihood_2D(),
    posterior_2D()
])
@pytest.mark.parametrize("log_scale", [True, False])
def test_plot_density(density, log_scale):
    """ Test the 1D and 2D density plot. """
    if density.dim == 1:
        plot_1D_density(density, -1, 1, color='green', log_scale=log_scale)
    elif density.dim == 2:
        plot_2D_density(
            density, -3, 3, -3, 3, 60, 60, cmap='gray', log_scale=log_scale)
    else:
        raise ValueError("Density must be 1D or 2D.")
    
def count_nonzero_threshold():
    threshold = 1e-8
    eps_min = 1e-1
    eps_plus = 1e1
    test_array = np.array([0, -eps_min*threshold, -threshold, -eps_plus*threshold, eps_min*threshold, threshold, eps_plus*threshold])

    nonzeros = count_nonzero(test_array, threshold)
    assert(nonzeros == 4)


def count_within_bounds_exceptions():
    test_array = np.array([-1.0, -0.5, 0.0, 0.5, 1.0])
    lower = -np.ones_like(test_array)
    upper = np.ones_like(test_array)
    
    within_bounds = count_within_bounds(test_array, lower, upper, exception=0.0)
    assert(within_bounds == 2)

def count_constant_components_1d_threshold():
    threshold = 1e-4
    test_array = np.array([0, 0, 1.0 - 0.5*threshold, 1.0, 1.0+0.5*threshold, 1.5, 2, 2+0.5*threshold, 2+threshold, 0, 0])

    components = count_constant_components_1D(test_array, threshold)
    assert(components == 5)

    components = count_constant_components_1D(test_array, threshold, lower = 0, upper = 2)
    assert(components == 2)

    components = count_constant_components_1D(test_array, threshold, exception = 0)
    assert(components == 3)

def count_constant_components_2d_threshold():
    threshold = 1e-4
    test_array = np.array([[0, 0                , 1 - 0.25*threshold],
                           [0, 1                , 1 + 0.25*threshold],
                           [2, 2 - 0.5*threshold, 2]])

    components = count_constant_components_2D(test_array, threshold)
    assert(components == 3)

    components = count_constant_components_2D(test_array, threshold, lower = 0, upper = 2)
    assert(components == 1)

    
def piecewise_linear_1D_DoF_exceptions():
    threshold = 1e-4
    test_array = np.array([0, 0, 0, 0.5, 1.0, 1.5, 2.0, 2.0, 2.0, 3.0, 4.0])

    components = piecewise_linear_1D_DoF(test_array, threshold)
    assert(components == 5)

    components = piecewise_linear_1D_DoF(test_array, threshold, exception_zero = True)
    assert(components == 3)

    components = piecewise_linear_1D_DoF(test_array, threshold, exception_flat = True)
    assert(components == 3)


from cuqi.utilities import get_non_default_args
from cuqi.distribution import Distribution, Gaussian
from cuqi.solver import ProjectNonnegative, ProjectBox, ProximalL1
from cuqi.geometry import Continuous1D, Continuous2D, Image2D
from cuqi.operator import FirstOrderFiniteDifference, SecondOrderFiniteDifference, Operator

import numpy as np
import scipy.sparse as sparse
import scipy.optimize as spoptimize

from copy import copy


class RegularizedGaussian(Distribution):
    """ Implicit Regularized Gaussian.

    Defines a so-called implicit prior based on a Gaussian distribution with implicit regularization.
    The regularization can be defined in the form of a proximal operator or a projector. 
    Alternatively, preset constraints and regularization can be used.

    Precisely one of proximal, projector, constraint or regularization needs to be provided. Otherwise, an error is raised.

    Can be used as a prior in a posterior which can be sampled with the RegularizedLinearRTO sampler.


    For more details on implicit regularized Gaussian see the following paper:

    [1] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Sparse Bayesian inference with regularized
    Gaussian distributions." Inverse Problems 39.11 (2023): 115004.

    Parameters
    ----------
    mean
        See :class:`~cuqi.distribution.Gaussian` for details.

    cov
        See :class:`~cuqi.distribution.Gaussian` for details.

    prec
        See :class:`~cuqi.distribution.Gaussian` for details.

    sqrtcov
        See :class:`~cuqi.distribution.Gaussian` for details.

    sqrtprec
        See :class:`~cuqi.distribution.Gaussian` for details.

    proximal : callable f(x, scale), list of tuples (callable proximal operator of f_i, linear operator L_i) or None
        If callable:
            Euclidean proximal operator f of the regularization function g, that is, a solver for the optimization problem
            min_z 0.5||x-z||_2^2+scale*g(x).
        If list of tuples (callable proximal operator of f_i, linear operator L_i):
            Each callable proximal operator of f_i accepts two arguments (x, p) and should return the minimizer of p/2||x-z||^2 + f(x) over z for some f.
            Each linear operator needs to have the '__matmul__', 'T' and 'shape' attributes;
            this includes numpy.ndarray, scipy.sparse.sparray, scipy.sparse.linalg.LinearOperator and cuqi.operator.Operator.
            The corresponding regularization takes the form
                sum_i f_i(L_i x),
            where the sum ranges from 1 to an arbitrary n.

    projector : callable f(x) or None
        Euclidean projection onto the constraint C, that is, a solver for the optimization problem
        min_(z in C) 0.5||x-z||_2^2.

    constraint : string or None
        Preset constraints that generate the corresponding proximal parameter. Required for use in Gibbs. For any geometry the following can be chosen:
        - "nonnegativity"
        - "box", the following additional parameters can be passed:
            lower_bound : array_like or None
                Lower bound of box, defaults to zero
            upper_bound : array_like
                Upper bound of box, defaults to one
        Additionally, for Continuous1D geometry the following options can be chosen:
        - "increasing"
        - "decreasing"
        - "convex"
        - "concave"

    regularization : string or None
        Preset regularization that generate the corresponding proximal parameter. Can be set to "l1" or 'tv'. Required for use in Gibbs in future update.
        For "l1" or "tv", the following additional parameters can be passed:
            strength : scalar
                Regularization parameter, i.e., strength*||Lx||_1, defaults to one

    """
        
    def __init__(self, mean=None, cov=None, prec=None, sqrtcov=None, sqrtprec=None, proximal = None, projector = None, constraint = None, regularization = None, **kwargs):
        
        # Store regularization parameters and remove them from kwargs passed to Gaussian
        optional_regularization_parameters = {
            "lower_bound" : kwargs.pop("lower_bound", None), # Takes default of ProjectBox if None
            "upper_bound" : kwargs.pop("upper_bound", None), # Takes default of ProjectBox if None
            "strength" : kwargs.pop("strength", 1)
        }
        
        # We init the underlying Gaussian first for geometry and dimensionality handling
        self._gaussian = Gaussian(mean=mean, cov=cov, prec=prec, sqrtcov=sqrtcov, sqrtprec=sqrtprec, **kwargs)
        kwargs.pop("geometry", None)

        # Init from abstract distribution class
        super().__init__(**kwargs)

        self._force_list = False
        self._parse_regularization_input_arguments(proximal, projector, constraint, regularization, optional_regularization_parameters)

    def _parse_regularization_input_arguments(self, proximal, projector, constraint, regularization, optional_regularization_parameters):
        """ Parse regularization input arguments with guarding statements and store internal states """

        # Guards checking whether the regularization inputs are valid
        if (proximal is not None) + (projector is not None) + max((constraint is not None), (regularization is not None)) == 0:
            raise ValueError("At least some constraint or regularization has to be specified for RegularizedGaussian")
            
        if (proximal is not None) + (projector is not None) == 2:
            raise ValueError("Only one of proximal or projector can be used.")

        if (proximal is not None) + (projector is not None) + max((constraint is not None), (regularization is not None)) > 1:
            raise ValueError("User-defined proximals and projectors cannot be combined with pre-defined constraints and regularization.")

        # Branch between user-defined and preset
        if (proximal is not None) + (projector is not None) >= 1:
            self._parse_user_specified_input(proximal, projector)
        else:
            # Set constraint and regularization presets for use with Gibbs
            self._preset = {"constraint": None,
                            "regularization": None}

            self._parse_preset_constraint_input(constraint, optional_regularization_parameters)
            self._parse_preset_regularization_input(regularization, optional_regularization_parameters)

            # Merge
            self._merge_predefined_option()

    def _parse_user_specified_input(self, proximal, projector):
        # Guard for checking partial validy of proximals or projectors
        if proximal is not None:
            if callable(proximal):
                if len(get_non_default_args(proximal)) != 2:
                    raise ValueError("Proximal should take 2 arguments.")
            elif isinstance(proximal, list):
                for val in proximal:
                    if len(val) != 2:
                        raise ValueError("Each value in the proximal list needs to consistent of two elements: a proximal operator and a linear operator.")
                    if callable(val[0]):
                        if len(get_non_default_args(val[0])) != 2:
                            raise ValueError("Proximal should take 2 arguments.")
                    else:
                        raise ValueError("Proximal operators need to be callable.")
                    if not (hasattr(val[1], '__matmul__') and hasattr(val[1], 'T') and hasattr(val[1], 'shape')):
                        raise ValueError("Linear operator not supported, must have '__matmul__', 'T' and 'shape' attributes.")
            else:
                raise ValueError("Proximal needs to be callable or a list. See documentation.")
            
        if projector is not None:
            if callable(projector):
                if len(get_non_default_args(projector)) != 1:
                    raise ValueError("Projector should take 1 argument.")
            else:
                raise ValueError("Projector needs to be callable")
            
        # Set user-defined proximals or projectors
        if proximal is not None:
            self._preset = None
            self._proximal = proximal
            return
        
        if projector is not None:
            self._preset = None
            self._proximal = lambda z, gamma: projector(z)
            return

    def _parse_preset_constraint_input(self, constraint, optional_regularization_parameters):
        # Create data for constraints
        self._constraint_prox = None
        self._constraint_oper = None
        if constraint is not None:
            if not isinstance(constraint, str):
                raise ValueError("Constraint needs to be specified as a string.")
            
            c_lower = constraint.lower()
            if c_lower == "nonnegativity":
                self._constraint_prox = lambda z, gamma: ProjectNonnegative(z)
                self._box_bounds = (np.ones(self.dim)*0, np.ones(self.dim)*np.inf)
                self._preset["constraint"] = "nonnegativity"
            elif c_lower == "box":
                _box_lower = optional_regularization_parameters["lower_bound"]
                _box_upper = optional_regularization_parameters["upper_bound"]
                self._proximal = lambda z, _: ProjectBox(z, _box_lower, _box_upper)
                self._box_bounds = (np.ones(self.dim)*_box_lower, np.ones(self.dim)*_box_upper)
                self._preset["constraint"] = "box"
            elif c_lower == "increasing":
                if not isinstance(self.geometry, Continuous1D):
                    raise ValueError("Geometry not supported for " + c_lower)
                if hasattr(spoptimize, 'isotonic_regression'):
                    self._constraint_prox = lambda z, _: spoptimize.isotonic_regression(z, increasing=True).x
                else:
                    raise AttributeError(f"The function 'isotonic_regression' does not exist in scipy.optimize. Installed scipy version: {spoptimize.__version__}. You need to install a scipy >= 1.12.0")
                self._preset["constraint"] = "increasing"
            elif c_lower == "decreasing":
                if not isinstance(self.geometry, Continuous1D):
                    raise ValueError("Geometry not supported for " + c_lower)
                if hasattr(spoptimize, 'isotonic_regression'):
                    self._constraint_prox = lambda z, _: spoptimize.isotonic_regression(z, increasing=False).x
                else:
                    raise AttributeError(f"The function 'isotonic_regression' does not exist in scipy.optimize. Installed scipy version: {spoptimize.__version__}. You need to install a scipy >= 1.12.0")
                self._preset["constraint"] = "decreasing"
            elif c_lower == "convex":
                if not isinstance(self.geometry, Continuous1D):
                    raise ValueError("Geometry not supported for " + c_lower)
                self._constraint_prox = lambda z, _: -ProjectNonnegative(-z)
                self._constraint_oper = SecondOrderFiniteDifference(self.geometry.fun_shape, bc_type='neumann')
                self._preset["constraint"] = "convex"
            elif c_lower == "concave":
                if not isinstance(self.geometry, Continuous1D):
                    raise ValueError("Geometry not supported for " + c_lower)
                self._constraint_prox = lambda z, _: ProjectNonnegative(z)
                self._constraint_oper = SecondOrderFiniteDifference(self.geometry.fun_shape, bc_type='neumann')
                self._preset["constraint"] = "concave"
            else:
                raise ValueError("Constraint not supported.")

    def _parse_preset_regularization_input(self, regularization, optional_regularization_parameters):
        # Create data for regularization
        self._regularization_prox = None
        self._regularization_oper = None
        if regularization is not None:
            if not isinstance(regularization, str):
                raise ValueError("Regularization needs to be specified as a string.")
                
            self._strength = optional_regularization_parameters["strength"]
            r_lower = regularization.lower()
            if r_lower == "l1":
                self._regularization_prox = lambda z, gamma: ProximalL1(z, gamma*self._strength)
                self._preset["regularization"] = "l1"
            elif r_lower == "tv":
                # Store the transformation to reuse when modifying the strength
                if not isinstance(self.geometry, (Continuous1D, Continuous2D, Image2D)):
                    raise ValueError("Geometry not supported for total variation")
                self._regularization_prox = lambda z, gamma: ProximalL1(z, gamma*self._strength)
                self._regularization_oper = FirstOrderFiniteDifference(self.geometry.fun_shape, bc_type='neumann')
                self._preset["regularization"] = "tv"
            else:
                raise ValueError("Regularization not supported.")
    
    def _merge_predefined_option(self):
        # Check whether it is a single proximal and hence FISTA could be used in RegularizedLinearRTO 
        if ((not self._force_list) and
            ((self._constraint_prox is not None) + (self._regularization_prox is not None) == 1) and
            ((self._constraint_oper is not None) + (self._regularization_oper is not None) == 0)):
                if self._constraint_prox is not None:
                    self._proximal = self._constraint_prox
                else:
                    self._proximal = self._regularization_prox 
                return

        # Merge regularization choices in list for use in ADMM by RegularizedLinearRTO
        self._proximal = []
        if self._constraint_prox is not None:
            self._proximal += [(self._constraint_prox, self._constraint_oper if self._constraint_oper is not None else sparse.eye(self.geometry.par_dim))]
        if self._regularization_prox is not None:
            self._proximal += [(self._regularization_prox, self._regularization_oper if self._regularization_oper is not None else sparse.eye(self.geometry.par_dim))]

    @property
    def transformation(self):
        return self._transformation
    
    @property
    def strength(self):
        return self._strength
        
    @strength.setter
    def strength(self, value):
        if self._preset is None or self._preset["regularization"] is None:
            raise TypeError("Strength is only used when the regularization is set to l1 or TV.")

        self._strength = value
        if self._preset["regularization"] in ["l1", "tv"]:        
            self._regularization_prox = lambda z, gamma: ProximalL1(z, gamma*self._strength)

        # Create new list of proximals based on updated regularization
        self._merge_predefined_option()

    # This is a getter only attribute for the underlying Gaussian
    # It also ensures that the name of the underlying Gaussian
    # matches the name of the implicit regularized Gaussian
    @property
    def gaussian(self):
        if self._name is not None:
            self._gaussian._name = self._name
        return self._gaussian
    
    @property
    def proximal(self):
        return self._proximal
    
    @proximal.setter
    def proximal(self, value):
        if callable(value):
            if len(get_non_default_args(value)) != 2:
                raise ValueError("Proximal should take 2 arguments.")
        elif isinstance(value, list):
            for (prox, op) in value:
                if len(get_non_default_args(prox)) != 2:
                    raise ValueError("Proximal should take 2 arguments.")
                if op.shape[1] != self.geometry.par_dim:
                    raise ValueError("Incorrect shape of linear operator in proximal list.")
        else:
            raise ValueError("Proximal needs to be callable or a list. See documentation.")
        
        self._proximal = value

        # For all the presets, self._proximal is set directly, 
        self._preset = None
            
    @property
    def preset(self):
        return self._preset

    def logpdf(self, x):
        return np.nan
        #raise ValueError(
        #    f"The logpdf of a implicit regularized Gaussian is not be defined.")
        
    def _sample(self, N, rng=None):
        raise ValueError(
            "Cannot be sampled from.")
  
    @staticmethod
    def constraint_options():
        return ["nonnegativity", "box", "increasing", "decreasing", "convex", "concave"]

    @staticmethod
    def regularization_options():
        return ["l1", "tv"]


    # --- Defer behavior of the underlying Gaussian --- #
    @property
    def geometry(self):
        return self.gaussian.geometry
    
    @geometry.setter
    def geometry(self, value):
        self.gaussian.geometry = value
    
    @property
    def mean(self):
        return self.gaussian.mean
    
    @mean.setter
    def mean(self, value):
        self.gaussian.mean = value
    
    @property
    def cov(self):
        return self.gaussian.cov
    
    @cov.setter
    def cov(self, value):
        self.gaussian.cov = value
    
    @property
    def prec(self):
        return self.gaussian.prec
    
    @prec.setter
    def prec(self, value):
        self.gaussian.prec = value
    
    @property
    def sqrtprec(self):
        return self.gaussian.sqrtprec
    
    @sqrtprec.setter
    def sqrtprec(self, value):
        self.gaussian.sqrtprec = value
    
    @property
    def sqrtcov(self):
        return self.gaussian.sqrtcov
    
    @sqrtcov.setter
    def sqrtcov(self, value):
        self.gaussian.sqrtcov = value     
    
    def get_mutable_variables(self):
        mutable_vars = self.gaussian.get_mutable_variables().copy()
        if self.preset is not None and self.preset['regularization'] in ["l1", "tv"]:
            mutable_vars += ["strength"]
        return mutable_vars
    
    def _make_copy(self):
        """ Returns a shallow copy of the density keeping a pointer to the original. """
        # Using deepcopy would also copy the underlying geometry, which causes a crash because geometries won't match anymore.
        new_density = copy(self)
        new_density._gaussian = copy(new_density._gaussian)
        new_density._original_density = self
        return new_density


class ConstrainedGaussian(RegularizedGaussian):
    """ Implicit Constrained Gaussian.

    Defines a so-called implicit prior based on a Gaussian distribution with implicit constraints.
    The constraint can be defined as a preset or in the form of a projector. 

    Precisely one of projector or constraint needs to be provided. Otherwise, an error is raised.

    Can be used as a prior in a posterior which can be sampled with the RegularizedLinearRTO sampler.

    Alias for :class:`~cuqi.implicitprior.RegularizedGaussian` with only constraints available.

    For more details on implicit regularized Gaussian see the following paper:

    [1] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Sparse Bayesian inference with regularized
    Gaussian distributions." Inverse Problems 39.11 (2023): 115004.

    Parameters
    ----------
    mean
        See :class:`~cuqi.distribution.Gaussian` for details.

    cov
        See :class:`~cuqi.distribution.Gaussian` for details.

    prec
        See :class:`~cuqi.distribution.Gaussian` for details.

    sqrtcov
        See :class:`~cuqi.distribution.Gaussian` for details.

    sqrtprec
        See :class:`~cuqi.distribution.Gaussian` for details.

    projector : callable f(x) or None
        Euclidean projection onto the constraint C, that is, a solver for the optimization problem
        min_(z in C) 0.5||x-z||_2^2.

    constraint : string or None
        Preset constraints that generate the corresponding proximal parameter. Can be set to "nonnegativity" and "box". Required for use in Gibbs.
        For "box", the following additional parameters can be passed:
            lower_bound : array_like or None
                Lower bound of box, defaults to zero
            upper_bound : array_like
                Upper bound of box, defaults to one

    """
    def __init__(self, mean=None, cov=None, prec=None, sqrtcov=None,sqrtprec=None, projector=None, constraint=None, **kwargs):
        super().__init__(mean=mean, cov=cov, prec=prec, sqrtcov=sqrtcov, sqrtprec=sqrtprec, projector=projector, constraint=constraint, **kwargs)

        
class NonnegativeGaussian(RegularizedGaussian):
    """ Implicit Nonnegative Gaussian.

    Defines a so-called implicit prior based on a Gaussian distribution with implicit nonnegativity constraints.

    Can be used as a prior in a posterior which can be sampled with the RegularizedLinearRTO sampler.

    Alias for :class:`~cuqi.implicitprior.RegularizedGaussian` with only nonnegativity constraints.

    For more details on implicit regularized Gaussian see the following paper:

    [1] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Sparse Bayesian inference with regularized
    Gaussian distributions." Inverse Problems 39.11 (2023): 115004.

    Parameters
    ----------
    mean
        See :class:`~cuqi.distribution.Gaussian` for details.

    cov
        See :class:`~cuqi.distribution.Gaussian` for details.

    prec
        See :class:`~cuqi.distribution.Gaussian` for details.

    sqrtcov
        See :class:`~cuqi.distribution.Gaussian` for details.

    sqrtprec
        See :class:`~cuqi.distribution.Gaussian` for details.

    """
    def __init__(self, mean=None, cov=None, prec=None, sqrtcov=None,sqrtprec=None, **kwargs):
        super().__init__(mean=mean, cov=cov, prec=prec, sqrtcov=sqrtcov, sqrtprec=sqrtprec, constraint="nonnegativity", **kwargs)


import numpy as np
import scipy.stats as sps
import scipy.special as special
from cuqi.distribution import Distribution
from cuqi.utilities import force_ndarray

class ModifiedHalfNormal(Distribution):
    """
    Represents a modified half-normal (MHN) distribution, a three-parameter family of distributions generalizing the Gamma distribution.
    The distribution is continuous with pdf

    .. math::
    
        f(x; \\alpha, \\beta, \\gamma) \propto x^{(\\alpha-1)} * \exp(-\\beta * x^2 + \\gamma * x)

    The MHN generalizes the half-normal distribution, because
    :math:`f(x; 1, \\beta, 0) \propto \exp(-\\beta * x^2)`

    The MHN generalizes the gamma distribution because
    :math:`f(x; \\alpha, 0, -\\gamma) \propto x^{(\\alpha-1)} * \exp(- \\gamma * x)`

    Reference:
    [1] Sun, et al. "The Modified-Half-Normal distribution: Properties and an efficient sampling scheme." Communications in Statistics-Theory and Methods
    
    Parameters
    ----------
    alpha : float or array_like
        The polynomial exponent parameter :math:`\\alpha` of the MHN distribution. Must be positive.

    beta : float or array_like
        The quadratic exponential parameter :math:`\\beta` of the MHN distribution. Must be positive.

    gamma : float or array_like
        The linear exponential parameter :math:`\\gamma` of the MHN distribution.

    """
    def __init__(self, alpha=None, beta=None, gamma=None, is_symmetric=False, **kwargs):
        # Init from abstract distribution class
        super().__init__(is_symmetric=is_symmetric, **kwargs) 

        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    @property
    def alpha(self):
        """ The polynomial exponent parameter of the MHN distribution. Must be positive. """
        return self._alpha
    
    @alpha.setter
    def alpha(self, value):
        self._alpha = force_ndarray(value, flatten=True)

    @property
    def beta(self):
        """ The quadratic exponential parameter of the MHN distribution. Must be positive. """
        return self._beta
    
    @beta.setter
    def beta(self, value):
        self._beta = force_ndarray(value, flatten=True)

    @property
    def gamma(self):
        """ The linear exponential parameter of the MHN distribution. """
        return self._gamma
    
    @gamma.setter
    def gamma(self, value):
        self._gamma = force_ndarray(value, flatten=True)

    def logpdf(self, x): # Unnormalized
        return np.sum((self.alpha - 1)*np.log(x) - self.beta * x * x + self.gamma * x)

    def _gradient_scalar(self, val):
        if val <= 0.0:
            return np.nan
        return (self.alpha - 1)/val - 2*self.beta*val + self.gamma

    def _gradient(self, val, *args, **kwargs):
        return np.array([self._gradient_scalar(v) for v in val])

    def _MHN_sample_gamma_proposal(self, alpha, beta, gamma, rng, delta=None):
        """
            Sample from a modified half-normal distribution using a Gamma distribution proposal.
        """
        if delta is None:
            delta = beta + (gamma*gamma - gamma*np.sqrt(gamma*gamma + 8*beta*alpha))/(4*alpha)
            
        while True:
            T = rng.gamma(alpha/2, 1.0/delta)
            X = np.sqrt(T)
            U = rng.uniform()
            if X > 0 and np.log(U) < -(beta-delta)*T + gamma*X - gamma*gamma/(4*(beta-delta)):
                return X
            
    def _MHN_sample_normal_proposal(self, alpha, beta, gamma, mu, rng):
        """
            Sample from a modified half-normal distribution using a Normal/Gaussian distribution proposal.
        """
        if mu is None:
            mu = (gamma + np.sqrt(gamma*gamma + 8*beta*(alpha - 1)))/(4*beta)    
        
        while True:
            X = rng.normal(mu, np.sqrt(0.5/beta))
            U = rng.uniform()
            if X > 0 and np.log(U) < (alpha-1)*np.log(X) - np.log(mu) + (2*beta*mu-gamma)*(mu-X):
                return X

    def _MHN_sample_positive_gamma_1(self, alpha, beta, gamma, rng):
        """
            Sample from a modified half-normal distribution, assuming alpha is greater than one and gamma is positive.
        """
        if gamma <= 0.0:
            raise ValueError("gamma needs to be positive")
            
        if alpha <= 1.0:
            raise ValueError("alpha needs to be greater than 1.0")

        # Decide whether to use Normal or sqrt(Gamma) proposals for acceptance-rejectance scheme
        mu = (gamma + np.sqrt(gamma*gamma + 8*beta*(alpha - 1)))/(4*beta)
        K1  = 2*np.sqrt(np.pi)
        K1 *= np.power((np.sqrt(beta)*(alpha-1))/(2*beta*mu-gamma), alpha - 1)
        K1 *= np.exp(-(alpha-1)+beta*mu*mu)
        
        delta = beta + (gamma*gamma - gamma*np.sqrt(gamma*gamma + 8*beta*alpha))/(4*alpha)
        K2  = np.power(beta/delta, 0.5*alpha)
        K2 *= special.gamma(alpha/2.0)
        K2 *= np.exp(gamma*gamma/(4*(beta-delta)))

        if K2 > K1: # Use normal proposal
            return self._MHN_sample_normal_proposal(alpha, beta, gamma, mu, rng)
        else: # Use sqrt(gamma) proposal
            return self._MHN_sample_gamma_proposal(alpha, beta, gamma, rng, delta)

    def _MHN_sample_negative_gamma(self, alpha, beta, gamma, rng, m=None):
        """
            Sample from a modified half-normal distribution, assuming gamma is negative.
            The argument 'm' is the matching point, see Algorithm 3 from [1] for details.
        """
        if gamma > 0.0:
            raise ValueError("gamma needs to be negative")
            
        if m is None:
            if alpha <= 1.0:
                m = 1.0
            else:
                m = "mode"
        
        # The acceptance rate of this choice is at least 0.5*sqrt(2) approx 70.7 percent, according to Theorem 4 from [1].
        if isinstance(m, str) and m.lower() == "mode":
            m = (gamma + np.sqrt(gamma*gamma + 8*beta*alpha))/(4*beta)

        while True:
            val1 = (beta*m-gamma)/(2*beta*m-gamma)
            val2 = m*(beta*m-gamma)
            T = rng.gamma(alpha*val1, 1.0/val2)
            X = m*np.power(T,val1)
            U = rng.uniform()
            if np.log(U) < val2*T-beta*X*X+gamma*X:
                return X
            
    def _MHN_sample(self, alpha, beta, gamma, m=None, rng=None):
        """
        Sample from a modified half-normal distribution using an algorithm from [1].
        """
        if rng == None:
            rng = np.random

        if gamma <= 0.0:
            return self._MHN_sample_negative_gamma(alpha, beta, gamma, m=m, rng=rng)
        
        if alpha > 1:
            return self._MHN_sample_positive_gamma_1(alpha, beta, gamma, rng=rng)
        
        return self._MHN_sample_gamma_proposal(alpha, beta, gamma, rng=rng)

    def _sample(self, N, rng=None):
        if hasattr(self.alpha, '__getitem__'):
            return np.array([[self._MHN_sample(self.alpha[i], self.beta[i], self.gamma[i], rng=rng) for i in range(len(self.alpha))] for _ in range(N)])
        else:
            return np.array([self._MHN_sample(self.alpha, self.beta, self.gamma, rng=rng) for i in range(N)])

            